import logging
from pathlib import Path

from olefile import OleFileIO, isOleFile
import numpy as np
from oxdls import OMEXML

from . import txrm_wrapper

class Inspector(OleFileIO):

    def __init__(self, filename):
        self.ole = None
        path = Path(filename)
        self._filename = path.name
        if path.exists() and isOleFile(filename):
            self.ole = super().__init__(filename)
            self._output_text = f"\n{self._filename}\n\n"

    def __enter__(self):
        self.ole = super().__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.ole is not None:
            super().__exit__(exc_type, exc_val, exc_tb)

    def inspect(self, extra=False):
        self._output_text = f"\n{self._filename}\n\n"

        if self.ole is not None:
            self._output_text += "{0} images of type '{1}' with dimensions: {2}\n".format(
                txrm_wrapper.extract_number_of_images(self.ole),
                txrm_wrapper.extract_image_dtype(self.ole, "ImageInfo")[0],
                ", ".join([str(i) for i in txrm_wrapper.extract_image_dims(self.ole)])
                )
            self._output_text += "Pixel size: {0}μm\n".format(
                txrm_wrapper.extract_pixel_size(self.ole))
            rows = (
                txrm_wrapper.read_imageinfo_as_int(self.ole, "MosiacRows")
                if self.ole.exists("ImageInfo/MosiacRows") else 0)
            columns = (
                txrm_wrapper.read_imageinfo_as_int(self.ole, "MosiacColumns")
                if self.ole.exists("ImageInfo/MosiacColumns") else 0)
            # Is a mosaic check:
            if rows > 1 or columns > 1:
                self._output_text += f"Is a mosaic of shape (rows x coumns): {rows}x{columns}\n"
            else:
                self._output_text += "Not a mosaic\n"
            if self.ole.exists("ReferenceData/Image"):
                self._output_text += "Reference of type '{0}' applied with dimensions: {1}\n".format(
                    txrm_wrapper.extract_image_dtype(self.ole, "ReferenceData")[0],
                    ", ".join([str(i) for i in txrm_wrapper.extract_ref_dims(self.ole)])
                    )
            else:
                self._output_text += "No reference applied\n"
            self._output_text += "\n"
            
            if extra:
                axis_dict = txrm_wrapper.get_axis_dict(self.ole)

                self._output_text += "Exposures (s): {0}\n\n".format(
                    ", ".join([str(i) for i in txrm_wrapper.extract_multiple_exposure_times(self.ole)]))
                self._output_text += "Xray magnification: {0}\n\n".format(
                    txrm_wrapper.extract_xray_magnification(self.ole))

                samplex = axis_dict.get(1)
                if samplex:
                    self._output_text += "{0} ({1}): {2}\n\n".format(
                        samplex[0], samplex[1],
                        ", ".join([str(i) for i in txrm_wrapper.extract_x_coords(self.ole)]))
                sampley = axis_dict.get(2)
                if sampley:
                    self._output_text += "{0} ({1}): {2}\n\n".format(
                        sampley[0], sampley[1],
                        ", ".join([str(i) for i in txrm_wrapper.extract_y_coords(self.ole)]))
                samplez = axis_dict.get(3)
                if samplez:
                    self._output_text += "{0} ({1}): {2}\n\n".format(
                        samplez[0], samplez[1],
                        ", ".join([str(i) for i in txrm_wrapper.extract_z_coords(self.ole)]))
                sampletheta = axis_dict.get(4)
                if sampletheta:
                    self._output_text += "{0} ({1}): {2}\n\n".format(
                        sampletheta[0], sampletheta[1],
                        ", ".join([str(i) for i in txrm_wrapper.extract_tilt_angles(self.ole)]))
                energy = axis_dict.get(13)
                if energy:
                    self._output_text += "{0} ({1}): {2}\n\n".format(
                        energy[0], energy[1],
                        ", ".join([str(i) for i in txrm_wrapper.extract_energies(self.ole)]))
                self._output_text += "Axis units: {0}\n".format(
                        ", ".join([" ".join(pair) for pair in axis_dict.values()]))
                self._output_text += "\n"
        else:
            self._output_text += "This is not a valid txrm/xrm file."

    def _get_stream_list(self):
        return ["/".join(stream) for stream in self.ole.listdir(streams=True, storages=False)]
            

    def list_streams(self):
        if self.ole is not None:
            for stream in self._get_stream_list():
                self._output_text += f"{stream}\n"

    def inspect_streams(self, *keys):
        if self.ole is not None:
            self._output_text += "Inspecting streams: " + ", ".join(keys) + "\n"
            try:
                for key in keys:
                    if self.ole.exists(key):
                        stream = self.ole.openstream(key)
                        stream_binary = stream.read()

                        self._output_text += f"\n\n{key}:\n"
                        self._output_text += f"\tbinary:           \t{stream_binary}\n"
                        self._output_text += f"\thex binary:       \t{stream_binary.hex()}\n"

                        for buff_type in txrm_wrapper.data_type_dict.values():
                            if buff_type[1] is not None:
                                try:
                                    values = ",".join(
                                        [str(i) for i in np.frombuffer(stream.getvalue(), buff_type[1]).tolist()])
                                    self._output_text += \
                                        f"\t{buff_type[0] + ':':18s}\t{values}\n"
                                except ValueError as e:
                                    self._output_text += f"\t{buff_type[0] + ':':18s}\t{e}\n"
                    else:
                        self._output_text += f"{key} does not exist\n"
            except Exception as e:
                self._output_text += f"\n{e}\n"

    def get_text(self):
        return self._output_text



import unittest
from unittest.mock import patch, MagicMock, ANY

import os
import sys
from pathlib import Path
from subprocess import Popen, PIPE, STDOUT

from txrm2tiff import __version__
from txrm2tiff import argparse_entrypoint


class TestCommandlineEntryPoint(unittest.TestCase):
# Before these running tests, the module must be installed in development mode using:
# `pip install --editable . --user` (from the project folder)
# it can be uninstalled at the end using:
# `pip uninstall txrm2tiff`
# IF TESTING ON WINDOWS: make sure the correct python installation is set as the default in PATH

    @classmethod
    def setUpClass(cls):
        # Set maximum difference string length to None (infinite)
        cls.maxDiff = None
        # Add local intall to path for testing with development install
        if os.name == "posix":
            os.environ["PATH"] += (
                os.pathsep
                + os.path.join(os.path.expanduser("~"), ".local", "bin")
                )
    
    @patch('argparse.ArgumentParser.print_help')
    def test_argparse_function(self, mocked_printhelp):
        sys.argv = ["txrm2tiff"]
        argparse_entrypoint.main()

        mocked_printhelp.assert_called()

    @patch('argparse.ArgumentParser.print_help')
    def test_argparse_function_setup_subpaser(self, mocked_printhelp):
        sys.argv = ["txrm2tiff", "setup"]
        argparse_entrypoint.main()

        mocked_printhelp.assert_called_once()

    def test_argparse_function_setup_windows_shortcut(self):
        with patch('txrm2tiff.shortcut_creator.create_Windows_shortcut', MagicMock()) as mocked_shortcut_creator:
            sys.argv = ["txrm2tiff", "setup", "-w"]
            argparse_entrypoint.main()
            
            mocked_shortcut_creator.assert_called_once()

    def test_argparse_function_with_args(self):
        with patch('txrm2tiff.run.run', MagicMock()) as mock_run:
            mock_run.return_value = None
            input_arg = "input_path"
            ref_arg = None
            data_type_arg = None
            sys.argv =["txrm2tiff", "--input", input_arg]
            argparse_entrypoint.main()

            mock_run.assert_called_once_with(input_path=input_arg, custom_reference=ref_arg, output_path=None, data_type=data_type_arg, ignore_reference=False, logging_level='info')

    def test_argparse_function_with_all_args(self):
        with patch('txrm2tiff.run.run', MagicMock()) as mock_run:
            mock_run.return_value = None
            input_arg = "input_path"
            ref_arg = "ref_path"
            data_type_arg = "uint16"
            sys.argv =["txrm2tiff", "--input", input_arg, "--reference", ref_arg, "--datatype", data_type_arg]
            argparse_entrypoint.main()

            mock_run.assert_called_once_with(input_path=input_arg, custom_reference=ref_arg, output_path=None, data_type=data_type_arg, ignore_reference=False, logging_level='info')

    def test_script_method(self):
        args = ["path_to/input.txrm"]
        run_args = ["txrm2tiff", "-i", args[0]]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertIn(f"No such file or directory: {args[0]}", stdout, msg=f"Actual stdout: {stdout}")

    def test_script_method_version_number(self):
        run_args = ["txrm2tiff", "--version"]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertEqual(f"txrm2tiff {__version__}", stdout, msg=f"Actual stdout: {stdout}")

    def test_script_method_setup_subparser(self):
        run_args = ["txrm2tiff", "setup"]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = stdout.replace("\r\n", " ").replace("\n", " ")
        self.assertIn("usage: txrm2tiff setup [-w] [-h]", stdout, msg=f"Actual stdout: {stdout}")

    def test_script_method_inspect_subparser(self):
        run_args = ["txrm2tiff", "inspect"]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = " ".join([text.strip() for text in stdout.replace("\r\n", " ").split('\n')])
        self.assertIn("txrm2tiff inspect [-i INPUT_PATH] [-e] [-l] [-s STREAMS [STREAMS ...]] [-h]", stdout, msg=f"Actual stdout: {stdout}")

    def test_module_without_arguments_returns_help(self):
        run_args = [sys.executable, "-m", "txrm2tiff"]
        path = os.environ["PATH"]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertIn("Converter of txrm/xrm files to OME tif/tiff files", stdout, msg=f"Actual stdout: {stdout}")

    def test_module_function_with_args(self):
        input_path = "input_file_path"
        custom_reference = "ref_path"
        output_path = None
        data_type = None
        ignore_reference = False
        logging_level = 1
        
        run_args = [sys.executable, "-m", "txrm2tiff", "--input", input_path, "--reference", custom_reference, "--set-logging", str(logging_level)]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = stdout.replace("\r\n", " ").replace("\n", " ")
        self.assertIn(
            f"Running with arguments: input_path={input_path}, custom_reference={custom_reference}, output_path={output_path}, data_type={data_type}, ignore_reference={ignore_reference}, logging_level={logging_level}",
            stdout,
            msg=f"Actual stdout: {stdout}")
        self.assertIn(f"No such file or directory: {input_path}", stdout, msg=f"Actual stdout: {stdout}")

    def test_module_method(self):
        input_path = "path_to/input.txrm"
        custom_reference = None
        output_path = None
        data_type = None
        ignore_reference = False
        logging_level = 1
        args = [input_path]
        run_args = [sys.executable, "-m", "txrm2tiff", "-i", args[0], "--set-logging", str(logging_level)]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = stdout.replace("\r\n", " ").replace("\n", " ")
        self.assertIn(
            f"Running with arguments: input_path={input_path}, custom_reference={custom_reference}, output_path={output_path}, data_type={data_type}, ignore_reference={ignore_reference}, logging_level={logging_level}",
            stdout,
            msg=f"Actual stdout: {stdout}")
        self.assertIn(f"No such file or directory: {input_path}", stdout, msg=f"Actual stdout: {stdout}")

    def test_module_method_version_number(self):
        run_args = [sys.executable, "-m", "txrm2tiff", "--version"]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertEqual(f"txrm2tiff {__version__}", stdout, msg=f"Actual stdout: {stdout}")

    def test_module_method_setup_subparser_without_arguments_returns_help(self):
        run_args = [sys.executable, "-m", "txrm2tiff", "setup"]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertIn(f"txrm2tiff setup [-w] [-h]", stdout, msg=f"Actual stdout: {stdout}")
